;    THIS SOURCE CODE IS SUPPLIED  ``AS IS'' WITHOUT WARRANTY OF ANY KIND, 
;    AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH 
;    (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL 
;    WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED
;    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND
;    ANY WARRANTIES OR NON INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY AND
;    RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR NOR
;    JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE FOR 
;    DAMAGES OF ANY KIND RESULTING FROM ITS USE.  Without limiting the 
;    generality of the foregoing, neither the author, nor JAIR, nor JAIR's
;    publishers and distributors, warrant that the Source Code will be 
;    error-free, will operate without interruption, or will meet the needs 
;    of the user.


; main
; Calls trace-theory for each example in the dataset.
;
; Examples should be in the file exs.lisp
; Theory should be in the file theory.lisp

(defun main ()
  (init)
  (setq *out-stream* (open "tgci1-output" :direction :output))
  (do ((index 0 (1+ index)))
      ((>= index *num-exs*) nil)
      (trace-theory *theory* index)
      (format *out-stream* "~%"))
  (close *out-stream*))


; trace-theory
;
; Calls appropriate subroutine based on the operator at the top
; of the theory.

(defun trace-theory (theory example-num)
  (cond ((null theory) (format t "WARNING: null theory~%"))
        ((equal (car theory) 'parts)
         (trace-theory-list (cddr theory) example-num))
        ((equal (car theory) 'not) 
         (trace-neg (cadr theory) (caddr theory) example-num))
        ((equal (car theory) 'and) 
         (trace-conj (cadr theory) (cddr theory) example-num))
        ((equal (car theory) 'or) 
         (trace-disj (cadr theory) (cddr theory) example-num))
        (t (leaf-value theory example-num))))


; trace-neg
;
; The NOT operator returns the negative of the value of its child.
; Print this feature if the output-flag is set.

(defun trace-neg (output-flag theory example-num)
  (let ((return-value (* -1.0 (trace-theory theory example-num))))
       (if output-flag (format *out-stream* "~A," return-value))
       return-value))


; trace-conj
;
; The AND operator returns the average of the values of its children.
; Print this feature if the output-flag is set.

(defun trace-conj (output-flag conj-list example-num)
  (let* ((result-list (trace-theory-list conj-list example-num))
         (return-value (/ (apply '+ result-list) (* 1.0 (length result-list)))))
        (if output-flag (format *out-stream* "~A," return-value))
        return-value))


; trace-disj
;
; The OR operator returns the maximum of the values of its children.
; Print this feature if the output-flag is set.

(defun trace-disj (output-flag disj-list example-num)
  (let* ((result-list (trace-theory-list disj-list example-num))
         (return-value (apply 'max result-list)))
        (if output-flag (format *out-stream* "~A," return-value))
        return-value))


; trace-theory-list
;
; Runs trace-theory on a list of theories.

(defun trace-theory-list (theory-list example-num)
  (do ((theory-list theory-list (cdr theory-list))
       (result-list nil))
      ((null theory-list) (reverse result-list))
      (setq result-list 
            (cons (trace-theory (car theory-list) example-num) result-list))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Utilities
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; leaf-value
;
; If an example matches the condition at a leaf of the theory, return
;	1.0; otherwise return -1.0

(defun leaf-value (leaf example-num)
  (cond ((equal (value leaf) (aref *ex-array* example-num (attribute leaf)))
         1)
        (t -1)))
             
(defun attribute (attribute-value)
   (car attribute-value))

(defun value (attribute-value)
   (cadr attribute-value))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Initialization
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun init ()
  (if (probe-file "tgci1-output") (delete-file "tgci1-output"))
  (load "exs")
  (load "theory")
  (setq *num-exs* (length *examples*))
  (setq *num-attr* (1- (length (car *examples*))))
  (setq *ex-array* (make-array (list *num-exs* (1+ *num-attr*))
                               :initial-contents *examples*))
)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Format of input files
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Example of exs.lisp

;  (setq *examples* '(
;    (+       t a c t a g c a a t a c g c t)
;    (+       t g c t a t c c t g a c a g t)
;    (+       g t a c t a g a g a a c t a g)
;    (+       a a t t g t g a t g t g t a t) 
;  ))

; The first item in each list is the examples class.  The remaining items
; are feature values.




; Example of theory.lisp

;(setq *theory*
; '(parts nil
;    (and t
;      (and t (or nil (28 c) (28 a)) (29 a) (30 g) (31 g) (32 t) 
;             (or nil (33 a) (33 g)) (34 a) (35 g) (36 t))
;      (not nil
;        (or t 
;          (and nil (28 t) (29 a) (30 a)) 
;          (and nil (28 t) (29 a) (30 g)) 
;          (and nil (28 t) (29 g) (30 a)) 
;          (and nil (27 t) (28 a) (29 a)) 
;          (and nil (27 t) (28 a) (29 g)) 
;          (and nil (27 t) (28 g) (29 a)) 
;          (and nil (26 t) (27 a) (28 a)) 
;          (and nil (26 t) (27 a) (28 g)) 
;          (and nil (26 t) (27 g) (28 a)) 
;        )
;      )
;    )
;    (and t
;      (and t (or nil (28 c) (28 t)) (29 a) (30 g) (31 g))
;      (and t 
;        (or nil (16 c) (16 t))
;        (or nil (17 c) (17 t))
;        (or nil (18 c) (18 t))
;        (or nil (19 c) (19 t))
;        (or nil (20 c) (20 t))
;        (or nil (21 c) (21 t))
;        (or nil (22 c) (22 t))
;        (or nil (23 c) (23 t))
;        (or nil (24 c) (24 t))
;        (or nil (25 c) (25 t))
;      )
;      (not nil
;        (or t 
;          (and nil (31 t) (32 a) (33 a)) 
;          (and nil (31 t) (32 a) (33 g)) 
;          (and nil (31 t) (32 g) (33 a)) 
;          (and nil (32 t) (33 a) (34 a)) 
;          (and nil (32 t) (33 a) (34 g)) 
;          (and nil (32 t) (33 g) (34 a)) 
;          (and nil (33 t) (34 a) (35 a)) 
;          (and nil (33 t) (34 a) (35 g)) 
;          (and nil (33 t) (34 g) (35 a)) 
;        )
;      )
;    )
;  )
;)

; Each leaf has the format (attribute value).   So (22 c) means that
;	attribute 22 should have the value 'c' to meet the condition.
;
; A theory has the format (operator output-flag child1 child2 child3 ...).
; 	A child may be a leaf or an entire theory.
;
; The outout-flag tells which nodes of the theories should have values
;	output for them and which should just be used in intermediate 
;	calculations.   't' means output a feature. 'nil' means just
;	use in intermediate calculations.
;
; The operator controls how the children will be combined.
;	AND - average the values of the children
;	OR  - take the MAX of the values of the children
;	NOT - negate the value of the child.  Setting the output-flag
;		of a NOT produces a redendant feature.
;	PARTS - this is simply a way of handling a theory with multiple
;		parts (such as 3 class domain with a theory for 2 of the
;		classes as shown above).    The output-flag is not used
;		for this operator but should be left there as a place holder.

